#
# Timer and related classes
# Gbor SEBESTYN <segabor@elte.hu>
#

class TEvent
    attr        :time
    attr        :period
    attr_writer :delegate
    def initialize(t, rep, del)
        @time = t
        @period = rep
        @delegate = del
    end
    
    def repeats?
        @period > 0
    end

    def adjust(t)
        if @period > 0 and @time < t
            q, r = (t - @time).divmod(@period)
            @time = @time + ( r > 0 ? ( (q+1)*@period) : q*@period)
        end
    end

    def secs
        @time.to_i
    end

    # Aaaaaaannndddd ... AAAACTION!!! :)    
    def fire
        @delegate.fire(self) if (not @delegate.nil?) and @delegate.methods.include? "fire"
    end
end


class TQueue
    attr :q

    def initialize
        @q = []
    end

    # Add a TEvent
    def <<(t)
        @q << t if (not t.nil?) and t.is_a? TEvent and (not @q.include? t) 
    end

    # Add a TEvent
    def add(t)
        self << t
    end

    # Remove TEvent(s) from queue
    def remove(t)
        if t.is_a? Array
            t.each do |t1|
                @q.delete t1
            end 
        else
            @q.delete t
        end
    end
    
    # Get Events before time 't'
    def eventsBefore(t)
        raise "TQueue.eventsBefore: invalid time #{t}" if t.nil? or not t.is_a? Time
        evs = []
        @q.each do |t1|
            # evs << t1 if (t.to_i % 3600 ) > (t1.secs % 3600)
            evs << t1 if t1.time < t
        end
        
        evs
    end

    # remove events belonging to delegate
    def rem_d(d)
        @q.delete_if do |t1|
            t1.delegate == d
        end
    end

    # Purge Events before time 't'
    def purgeBefore(t)
        raise "TQueue.purgeBefore: invalid time #{t}" if t.nil? or not t.is_a? Time
        @q.delete_if do |t1|
            # (t.to_i % 3600 ) > (t1.secs % 3600) and not t1.repeats?
            t1.time < t and not t1.repeats?
        end
    end
end


class Timer
    attr    :tq     # event queue
    attr    :t      # current time
    def initialize
        @tq = TQueue.new
        @t  = Time.now
    end

    # Add event to queue
    def add(t)
        @tq << t
    end

    # add a new event to queue
    def adde(h, m, r, d)
        ev = TEvent.new(self.time_hm(h,m,true), r , d)
        add(ev)
        ev
    end

    # Remove event
    def remove(t)
        @tq.remove(t)
    end
    
    # Remove events belonging to a delegate
    def rem_d(d)
        @tq.rem_d(d)
    end
    
    # Increase timer with s secs and fire events
    def +(s)
        # add seconds to the time
        @t += s
        
        # fire events
        evs = @tq.eventsBefore(@t)
        evs.each {|te| te.fire}
        
        # purge events
        @tq.purgeBefore(@t)
        
        # adjust periodic events
        evs.each {|te| te.adjust(@t)}

        self
    end
    
    # return the current time in HH:MM
    def dtime
        @t.hour * 60 + @t.min
    end
    
    # util
    def time_hm(h, m, step=false)
        t1 = Time.local(@t.year, @t.month, @t.day, h, m, 0)
        if t1 < @t and step == true
            t1 += 24*3600
        end
        t1
    end
    
    # is time around HH:MM
    def around(h, m, diff)
        t = time_hm(h, m)
        (t-@t).abs < diff
    end
    
    def to_s
        # @t.month.to_s + " " + @t.day.to_s + " " + @t.hour.to_s + ":" + @t.min.to_s
        @t.to_s
    end
end


#
# Test Module
#
module TimerTest
    class EvTest
        attr    :ev
    
        def addToTimer(t)
            @ev = TEvent.new(Time.now, 3, self)
            t.add(@ev)
        end
    
        def removeFromTimer(t)
            t.remove(@ev)
        end
    
        def fire(t)
            puts "I'm fired"
        end
    end
    
    def TimerTest::Test
        tm = Timer.new
        
        evt = EvTest.new
        
        evt.addToTimer(tm)
        20.times {|k|
            tm = tm + 2
            puts "#{k}: timer adjusted by 2"
        }
        evt.removeFromTimer(tm)
    end
end
